from scipy import optimize
import numpy as np
import logging
import pysand.exceptions as exc

logger = logging.getLogger(__name__)

# Constants
g = 9.80665  # Standard gravity [m/s^2]


def hydro(D, rho_l, mu_l, d_p, e=5e-5, rho_p=2650):
    """
    Equinor sand transport model for horizontal pipelines
    Based on T. Søntvedt (1995) and R. Schulkes (2002) work in Hydro
    :param D: Pipe diameter [m]
    :param rho_l: Liquid density [kg/m3]
    :param mu_l: Dynamic liquid viscosity [Pa.s]
    :param d_p: Sand grain diameter [mm]
    :param e: Roughness of pipe [m] Optional. Default value: 5e-5 m
    :param rho_p: Particle density [kg/m3], default = 2650 (quartz sand)
    :return: list of critical velocities for transitions from stationary to moving bed
    and from moving bed to dispersed transport [m/s]
    """

    # Input validation
    kwargs = {'D': D, 'rho_l': rho_l, 'mu_l': mu_l, 'd_p': d_p, 'e': e, 'rho_p': rho_p}
    if validate_inputs(**kwargs):
        return np.nan

    d_p = d_p / 1000  # from mm to m
    nu = mu_l / rho_l  # Liquid kinematic viscotity [m2/s]
    result = []

    def moody(ed, Re):
        '''
        Finds darcy friction factor by solving Colebrook equation (Moody diagram).
        Laminar and turbulent flow are correctly accounted for.
        :param ed: relative roughness = epsilon/diameter
        :param Re: Reynolds number
        :return: friction factor
        '''

        if Re < 0:
            logger.warning('Reynolds number = {}. Cannot be negative'.format(Re))
        if Re < 2000:
            f = 64 / Re  # Laminar flow
            return f
        if ed > 0.05:
            logger.warning('epsilon/diameter ratio = {}, is not on Moody chart'.format(ed))
        if Re < 4000:
            logger.info('Reynolds number, Re = {}, is in the transition range'.format(Re))

        # Use fzero to find f from the Colebrook equation.
        # colebrook is a function to evaluate F(f,e/d,Re)
        # fzero returns the value of f such that F(f,e/d/Re) = 0 (approximately)
        # fi = initial guess from Haaland equation
        # Iterations of fzero are terminated when f is known to within +/- 5e-6

        def colebrook(f, ed, Re):
            cb = 1.0 / f ** 0.5 + 2.0 * np.log10(ed / 3.7 + 2.51 / (Re * f ** 0.5))
            return cb

        fi = 1 / (1.8 * np.log10(6.9 / Re + (ed / 3.7) ** 1.11)) ** 2  # initial guess at f
        fzero = optimize.newton(colebrook, fi, args=(ed, Re), tol=5e-6)

        # sanity check:
        if fzero < 0:
            logger.warning('Friction factor = {}, cannot be negative'.format(fzero))
        return fzero

    def critical_u(u, m):
        Re = (u * D) / nu  # Reynolds number
        f_turb = 0.25 * moody(e / D, Re)  # Fanning friction factor
        t_w = .5 * f_turb * rho_l * u ** 2  # Wall shear stress [Pa]
        uf = (t_w / rho_l) ** 0.5  # Friction velocity
        ds = d_p * (uf / nu)  # Dimensionless particle diameter
        SH = t_w / ((rho_p - rho_l) * g * d_p)  # Shield number
        SH_bed = 0.42 / ((ds) ** 0.49)  # Critical Shield number between fixed bed and scouring
        SH_scouring = 1.12 / ((ds) ** 0.37)  # Critical Shield number between scouring and dispersed
        if m == 'bed':
            return SH - SH_bed
        if m == 'scouring':
            return SH - SH_scouring

    for i in ['bed', 'scouring']:
        vel = optimize.newton(critical_u, 2, args=(i,), tol=5e-6, maxiter=100)
        result.append(np.round(vel, 2))

    return result


def stokes(rho_m, mu_m, d_p, angle, rho_p=2650):
    """
    Sand lifting rate calculation model for vertical or deviated (up to ~80 deg) inclination
    Based on Stokes law for settling velocities, but corrected for turbulent flow around sand grains
    Mixture fluid properties and inclination at point of interest (e.g. down hole in a well)
    :param rho_m: Mixture density [kg/m3]
    :param mu_m: Mixture viscosity [kg/ms]
    :param d_p: Particle diameter [mm]
    :param angle: Inclination from vertical [deg]
    :param rho_p: Particle density [kg/m3], default = 2650 (quartz sand)
    :return: Particle settling velocity at point of interest [m/s]
    """

    # Input validation
    kwargs = {'rho_m': rho_m, 'mu_m': mu_m, 'd_p': d_p, 'angle': angle, 'rho_p': rho_p}
    if validate_inputs(**kwargs):
        return np.nan

    def f(v_est):
        Re = rho_m * d_p/1000 / mu_m * v_est
        if Re > 1899:
            f = .4
        else:
            f = 10 ** (1.413 - 0.923 * np.log10(Re) + 0.113 * np.log10(Re)**2)  # Polynomial friction factor
        A1 = 4 * g * d_p/1000 * (rho_p - rho_m) / (3 * rho_m * np.cos(np.deg2rad(angle)))
        v_c = (A1 / f) ** 0.5  # Critical velocity
        return abs(v_est - v_c)
    result = optimize.minimize_scalar(f, method='bounded', bounds=([0.001, 100]))
    if result.success:
        return np.round(result.x, 2)
    else:
        logger.warning('The model was not able to converge and returned nan.')
        return np.nan


def validate_inputs(**kwargs):

    for i in ['rho_m', 'rho_l', 'mu_m', 'mu_l', 'd_p', 'e', 'rho_p', 'D']:
        if i in kwargs:
            if not isinstance(kwargs[i], (float, int)) or np.isnan(kwargs[i]):
                raise exc.FunctionInputFail('{} is not a number'.format(i))
            if not kwargs[i] >= 0:
                logger.warning('The model got negative value(s) of {} and returned nan.'.format(i))
                return True

    if 'angle' in kwargs:
        if not isinstance(kwargs['angle'], (float, int)) or np.isnan(kwargs['angle']):
            raise exc.FunctionInputFail('angle is not a number')
        if (kwargs['angle'] < 0) or (kwargs['angle'] > 89):
            raise exc.FunctionInputFail('The inclination has to be positive and below 90 deg.')

    return



def critical_velocity(p1, p2, rho_1, rho_2):
    """
    Throttling velocity in the choke
    DNVGL RP-O501, August 2015 - chapter 4.12.3
    :param p1: Choke upstream pressure [bar]
    :param p2: Choke downstream pressure [bar]
    :param rho_1: Fluid mixture density upstream choke [kg/m3]
    :param rho_2: Fluid mixture density downstream choke [kg/m3]
    :return: Throttling velocity [m/s]
    """
    v_c = (4e5 * (p1 - p2) / (rho_1 + rho_2)) ** (1/2)
    return v_c


def min_choke(Qs, v_c):
    """
    Recommended minimum choke opening for plug/cage and cage/sleeve types of chokes
    DNVGL RP-O501, August 2015 - chapter 4.12.3
    :param Qs: Sand rate [g/s]
    :param v_c: Throttling velocity [m/s]
    :return: Minimum relative Cv [-]
    """
    if Qs < 0.01:
        if v_c < 100:
            cvr_min = 0.05
        else:
            cvr_min = 0.1
    elif 0.01 <= Qs < 0.1:
        if v_c < 50:
            cvr_min = 0.05
        elif 50 <= v_c < 100:
            cvr_min = 0.1
        else:
            cvr_min = 0.2
    else:
        if v_c < 50:
            cvr_min = 0.1
        else:
            cvr_min = 0.2
    return cvr_min



import numpy as np
import logging
import pysand.exceptions as exc

logger = logging.getLogger(__name__)
# Models from DNVGL RP-O501, equation references in parenthesis

def validate_fluid_props(**kwargs):
    """
        Validation of all input parameters that go into fluid properties models
    """
    for i in ['P', 'T', 'Qo', 'Qw', 'Qg', 'Z', 'D', 'rho_o', 'rho_w', 'MW', 'mu_o', 'mu_w', 'mu_g']:
        if i in kwargs:
            if kwargs[i] is None:
                raise exc.FunctionInputFail('No fluid properties are calculated due to missing {}'.format(i))
            if not isinstance(kwargs[i], (float, int)):
                raise exc.FunctionInputFail('{} is not a number'.format(i))
            if not kwargs[i] >= 0:
                logger.warning('The model has got negative value(s) of {} and returned nan.'.format(i))
                return True

def mix_velocity(P, T, Qo, Qw, Qg, Z, D):
    '''
    Blackoil mixture velocity model, based on DNVGL RP-O501, August 2015 edition
    :param P: Pressure [bar]
    :param T: Temperature [deg C]
    :param Qo: Oil rate [Sm3/d]
    :param Qw: Water rate [Sm3/d]
    :param Qg: Gas rate [Sm3/d]
    :param Z: Gas compressibility factor [-]
    :param D: Cross sectional diameter [m]
    :return: Mix velocity [m/s]
    '''

    kwargs = {'P': P, 'T': T, 'Qo': Qo, 'Qw': Qw, 'Qg': Qg, 'Z': Z, 'D': D}
    if validate_fluid_props(**kwargs):
        return np.nan

    T = T + 273.15
    # Constants
    P0 = 1.01325  # Pressure at std conditions [bar]
    T0 = 289.0  # Temperature at std conditions [K]

    # Calculations
    v_m = (Qo + Qw + Qg * Z * P0 * T / (P * T0)) / (np.pi / 4 * D ** 2) / 24 / 3600  # (4.14, 4.15, 4.16)
    return np.round(v_m, 2)


def mix_density(P, T, Qo, Qw, Qg, rho_o, rho_w, MW, Z):
    '''
    Blackoil mixture density calculator, based on DNVGL RP-O501, August 2015 edition
    :param P: Pressure [bar]
    :param T: Temperature [deg C]
    :param Qo: Oil rate [Sm3/d]
    :param Qw: Water rate [Sm3/d]
    :param Qg: Gas rate [Sm3/d]
    :param rho_o: Oil density at std conditions [kg/m3]
    :param rho_w: Water density at std conditions [kg/m3]
    :param MW: Gas molecular weight [kg/kmol]
    :param Z: Gas compressibility factor [-]
    :return: Mix density [kg/m3]
    '''

    kwargs = {'P': P, 'T': T, 'Qo': Qo, 'Qw': Qw, 'Qg': Qg, 'rho_o': rho_o, 'rho_w': rho_w, 'MW': MW, 'Z': Z}
    if validate_fluid_props(**kwargs):
        return np.nan

    T = T + 273.15
    # Constants
    P0 = 1.01325  # Pressure at std conditions [bar]
    T0 = 289.0  # Temperature at std conditions [K]
    R = 8314.0  # Universal gas constant [J/kgK]

    # Calculations
    rho_m = (Qo * rho_o + Qw * rho_w + Qg * P0 * MW / (R * T0) * 1e5) / (Qo + Qw + Qg * Z * P0 * T / (P * T0))  # (4.17)
    return np.round(rho_m, 2)


def mix_viscosity(P, T, Qo, Qw, Qg, mu_o, mu_w, mu_g, Z):
    '''
    Blackoil mixture viscosity calculator, based on DNVGL RP-O501, August 2015 edition
    Output viscosity units equals input units
    :param P: Pressure [bar]
    :param T: Temperature [deg C]
    :param Qo: Oil rate [Sm3/d]
    :param Qw: Water rate [Sm3/d]
    :param Qg: Gas rate [Sm3/d]
    :param mu_o: Oil viscosity
    :param mu_w: Water viscosity
    :param mu_g: Gas viscosity
    :param Z: Gas compressibility factor [-]
    :return: Mixture viscosity
    '''

    kwargs = {'P': P, 'T': T, 'Qo': Qo, 'Qw': Qw, 'Qg': Qg, 'mu_o': mu_o, 'mu_w': mu_w, 'mu_g': mu_g, 'Z': Z}
    if validate_fluid_props(**kwargs):
        return np.nan

    T = T + 273.15
    # Constants
    P0 = 1.01325  # Pressure at std conditions [bar]
    T0 = 289.0  # Temperature at std conditions [K]

    # Calculations
    mu_m = (mu_o + Qw/Qo * mu_w + Qg/Qo*P0*T*Z / (P * T0) * mu_g) / (1 + Qw/Qo + Qg/Qo*P0*T*Z / (P * T0))
    return np.round(mu_m, 6)


def liq_density(Qo, Qw, rho_o, rho_w):
    """
    Weighted average of oil and water densities
    :param Qo: Oil rate [Sm3/d]
    :param Qw: Water rate [Sm3/d]
    :param rho_o: Oil density at std conditions [kg/m3]
    :param rho_w: Water density at std conditions [kg/m3]
    :return: Liquid density
    """

    kwargs = {'Qo': Qo, 'Qw': Qw, 'rho_o': rho_o, 'rho_w': rho_w}
    if validate_fluid_props(**kwargs):
        return np.nan

    rho_l = (rho_o * Qo + rho_w * Qw) / (Qo + Qw)

    return rho_l


def liq_viscosity(Qo, Qw, mu_o, mu_w):
    """
    Weighted average of oil and water viscosities
    :param Qo: Oil rate [Sm3/d]
    :param Qw: Water rate [Sm3/d]
    :param mu_o: Oil viscosity
    :param mu_w: Water viscosity
    :return: Liquid viscosity
    """

    kwargs = {'Qo': Qo, 'Qw': Qw, 'mu_o': mu_o, 'mu_w': mu_w}
    if validate_fluid_props(**kwargs):
        return np.nan

    mu_l = (mu_o * Qo + mu_w * Qw) / (Qo + Qw)

    return mu_l




from scipy import interpolate
import numpy as np
import pysand.exceptions as exc
import logging

logger = logging.getLogger(__name__)

def validate_asd(**kwargs):
    """
    Validation of all input parameters that go into std_step_clampon, std_step_emerson and sand_rate functions;
    """
    for i in ['v_m', 'GLR', 'GOR']:
        if i in kwargs:
            if kwargs[i] is None:
                raise exc.FunctionInputFail('No calculation is done due to missing {}'.format(i))
            if not isinstance(kwargs[i], (float, int, np.integer)):
                raise exc.FunctionInputFail('{} is not a number'.format(i))
            if not kwargs[i] >= 0:
                logger.warning('The model has got negative value(s) of {} and returned nan.'.format(i))
                return True

def std_step_clampon(v_m, GLR):
    '''
    Standard step calculation for Clampon ASD's
    :param v_m: fluid mix velocity at ASD [m/s]
    :param GLR: Gas liquid ratio [Sm3/Sm3]
    :return: Standard step value
    '''

    kwargs = {'v_m': v_m, 'GLR': GLR}
    if validate_asd(**kwargs):
        return np.nan

    # Standard Step values from Equinor wiki, Jan 18:
    step_v_m = [0, 1, 2, 3, 4, 6, 8, 12, 16, 20, 24, 28, 32, 36, 40, 45, 50]
    step_oil = [0, 425, 1500, 2500, 3100, 4500, 5500, 8100, 11300, 14500, 17700, 20900, 24100, 27300, 30500, 34500, 38500]
    step_gas = [0, 525, 1500, 2500, 3200, 4800, 6500, 12320, 20740, 31500, 44892, 58284, 74664, 94032, 113400, 137610, 161820]
    # Linear interpolation functions for STEP
    f_oil = interpolate.interp1d(step_v_m, step_oil, fill_value='extrapolate')
    f_gas = interpolate.interp1d(step_v_m, step_gas, fill_value='extrapolate')

    if GLR < 150:
        step = f_oil(v_m)
    else:
        step = f_gas(v_m)

    return step


def std_step_emerson(v_m, GOR):
    '''
    Standard step calculation for Emerson ASD's
    :param v_m: fluid mix velocity at ASD [m/s]
    :param GOR: Gas oil ratio [Sm3/Sm3]
    :return: Standard step value
    '''

    kwargs = {'v_m': v_m, 'GOR': GOR}
    if validate_asd(**kwargs):
        return np.nan

    # Standard Step values from Equinor wiki, Jan 18:
    if GOR > 150:
        E, F, G, H = 3.2, 149.5, 4486.9, 360.8
    else:
        E, F, G, H = 50, 48, 200, 1000

    step = E * v_m ** 3 + F * v_m ** 2 + G * v_m + H

    return step


def sand_rate(raw, zero, step, exp=1):
    '''
    ASD sand rate calculation
    :param raw: raw value from ASD
    :param zero: Background noise zero value (no sand production)
    :param step: Sand noise
    :param exp: Linearity exponent (calibration constant)
    :return: Sand rate [g/s]
    '''

    for key, value in {'raw': raw, 'zero': zero, 'step': step, 'exp': exp}.items():
        if value is None:
            raise exc.FunctionInputFail('No calculation is done due to missing {}'.format(key))

    if raw > zero:
        try:
            Qs = (raw - zero)**exp / step
        except ZeroDivisionError:
            logger.warning('Step value equal to zero, sand rate set to NaN')
            Qs = np.nan
        else:
            if Qs < 0:
                logger.warning('Negative step. Sand rate set to NaN')
                Qs = np.nan
    else:
        Qs = 0

    return Qs